package com.brilliance.ccbank.srcipt.parser;

import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;

import com.brilliance.ccbank.srcipt.core.ScriptParserException;
import com.brilliance.ccbank.srcipt.core.ScriptType;
import com.brilliance.ccbank.srcipt.model.impl.EndDef;
import com.brilliance.ccbank.srcipt.model.impl.ScriptCallDef;
import com.brilliance.ccbank.srcipt.reader.ScriptReader;
import com.brilliance.eibs.core.exception.InterfaceException;
import com.brilliance.eibs.core.model.BreakDef;
import com.brilliance.eibs.core.model.IAbsDef;
import com.brilliance.eibs.core.model.IForeachDef;
import com.brilliance.eibs.core.model.IValidatorDef;
import com.brilliance.eibs.core.model.IWhileDef;
import com.brilliance.eibs.core.model.impl.ContinueDef;
import com.brilliance.eibs.core.model.impl.ForeachDef;
import com.brilliance.eibs.core.model.impl.IfDef;
import com.brilliance.eibs.core.model.impl.ValidatorDef;
import com.brilliance.eibs.core.model.impl.WhileDef;

public abstract class TokenParser
{
  protected static final int ARGS_MINNUM           = 1;
  protected static final int CONDITION_ARGS_MINNUM = 5;
  protected static final int CALL_ARGS_MINNUM      = 4;
  protected static final int VALIDATOR_ARGS_MINNUM = 4;
  public final static String CC               = "CC";//逻辑组件
  public final static String BC               = "BC";//流程组件
  
  private List<IAbsDef>      executeList;
  protected Logger             logger;
  protected ScriptReader       scriptReader;

  public TokenParser( ScriptReader sr, Logger log )
  {
    this.logger = log;
    this.scriptReader = sr;
    this.executeList = new LinkedList<IAbsDef>();
  }

  public List<IAbsDef> getExecuteList()
  {
    return executeList;
  }

  public void parse()
      throws Exception
  {
    if (!checkBeforeParse())
      return;
    if (scriptReader.isBegin())
    {
      this.executeList.add(scriptReader.getBeginDef());
      this.executeList.addAll(getExecuteList(scriptReader.getBeginDef()));
    }
    else
    {
      logger.warn("未解析到启动脚本命令行");
      throw new ScriptParserException("未解析到启动脚本命令行");
    }
  }

  private List<IAbsDef> getExecuteList(IAbsDef parentDef)
      throws Exception
  {
    List<IAbsDef> executeList = new LinkedList<IAbsDef>();
    while (scriptReader.hasNext())
    {
      String line = scriptReader.next();
      String[] args = elementsOfLine(line);
      if (ScriptType.CALL.eq(args[0]))
      {
        checkArgs(line, args, CALL_ARGS_MINNUM);
        executeList.add(getScriptCallDef(parentDef, args));
      }
      else if (ScriptType.IF.eq(args[0]))
      {
        checkArgs(line, args, CONDITION_ARGS_MINNUM);
        IfDef ifDef = new IfDef();
        ifDef.setParentDef(parentDef);
        String condition = args[1];
        String[] conditons=new String[args.length-1];
        System.arraycopy(args, 1, conditons, 0, conditons.length);
        condition = StringUtils.join(conditons, " ");
        ifDef.addConditionMap(condition, getExecuteList(ifDef));
        while (scriptReader.hasNext())
        {
          String subline = scriptReader.next();
          String[] subargs = elementsOfLine(subline);
          
          if (ScriptType.ELSEIF.eq(subargs[0]))
          {
            checkArgs(subline, subargs, CONDITION_ARGS_MINNUM);
            conditons=new String[subargs.length-1];
            System.arraycopy(subargs, 1, conditons, 0, conditons.length);
            condition = StringUtils.join(conditons, " ");
            if (!ifDef.getConditionMap().containsKey(condition))
              ifDef.addConditionMap(condition, getExecuteList(ifDef));
          }
          else if (ScriptType.ELSE.eq(subargs[0]))
          {
            ifDef.setElseCondition(getExecuteList(ifDef));
          }
          else
          {
            scriptReader.backWard(1);
            break;
          }
        }
        executeList.add(ifDef);
      }
      else if (ScriptType.VALIDATOR.eq(args[0]))
      {
        checkArgs(line, args, VALIDATOR_ARGS_MINNUM);
        executeList.add(getValidatorDef(parentDef, args));
      }
      else if (ScriptType.CONTINUE.eq(args[0]))
      {
        checkArgs(line, args, ARGS_MINNUM);
        if (nodeInLoop(parentDef))
        {
          ContinueDef continueDef = new ContinueDef();
          continueDef.setParentDef(parentDef);
          executeList.add(continueDef);
        }
        else
        {
          logger.warn("unexpected [continue] node under [" + args[0] + "].");
          throw new InterfaceException("00024", "unexpected [continue] node under [" + args[0] + "].");
        }
      }
      else if (ScriptType.BREAK.eq(args[0]))
      {
        checkArgs(line, args, ARGS_MINNUM);
        if (nodeInLoop(parentDef))
        {
          BreakDef breakDef = new BreakDef();
          breakDef.setParentDef(parentDef);
          executeList.add(breakDef);
        }
        else
        {
          logger.warn("unexpected [break] node under [" + args[0] + "].");
          throw new InterfaceException("00023", "unexpected [break] node under [" + args[0] + "].");
        }

      }
      else if (ScriptType.WHILE.eq(args[0]))
      {
        checkArgs(line, args, CONDITION_ARGS_MINNUM);
        IWhileDef whileDef = new WhileDef();
        whileDef.setCondition(args[1]);
        whileDef.setParentDef(parentDef);
        whileDef.setSubList(getExecuteList(whileDef));
        executeList.add(whileDef);
      }
      else if (ScriptType.isEnd(args[0]))
      {
        executeList.add(new EndDef(args[1]));
        break;
      }
      else
      {
        logger.warn("未知脚本命令[" + line + "]");
        throw new ScriptParserException("未知脚本命令[" + line + "]");
      }
    }
    return executeList;
  }

  private boolean nodeInLoop(IAbsDef parentDef)
  {
    return parentDef instanceof IForeachDef || parentDef instanceof IWhileDef || parentDef.subNodeOf(ForeachDef.class)
        || parentDef.subNodeOf(WhileDef.class);
  }

  private IAbsDef getValidatorDef(IAbsDef parentDef, String[] args)
  {
    IValidatorDef validatorDef = new ValidatorDef();
    validatorDef.setId("");
    validatorDef.setCondition(args[1]);
    validatorDef.setErrorCode(getAttribute(args, "ERRORCODE"));
    validatorDef.setDescription(getAttribute(args, "ERRORDESC"));
    validatorDef.setParentDef(parentDef);
    return validatorDef;
  }

  private String getAttribute(String[] args, String dest)
  {
    String rawrs = "";
    for (String arg : args)
    {
      arg = arg.trim().toUpperCase();
      if (arg.startsWith(dest))
      {
        rawrs = arg.split("=")[1];
        if (rawrs.startsWith("'") && rawrs.endsWith("'"))
          rawrs = rawrs.substring(1, rawrs.length() - 1);
        break;
      }
    }
    return rawrs;
  }

  /**
   * 检测此命令行经过切割后，是否满足解析的最低参数量。不满足则抛出异常。
   * @param line 命令行
   * @param args 切割后的命令行参数数组
   * @param minargsnum 参数数组长度最低要求
   * @throws Exception
   */
  private void checkArgs(String line, String[] args, int minargsnum)
      throws Exception
  {
    if (args.length < minargsnum)
    {
      logger.warn("脚本命令[" + line + "]缺少参数,最少需要" + minargsnum + "个参数,实际只有" + args.length + "个参数");
      throw new ScriptParserException("脚本命令[" + line + "]缺少参数,最少需要" + minargsnum + "个参数,实际只有" + args.length + "个参数");
    }

  }

  private IAbsDef getScriptCallDef(IAbsDef parentDef, String[] args)
  {
    ScriptCallDef scriptDef = new ScriptCallDef();
    scriptDef.setCompId(args[1]);
    scriptDef.setProcessnum(Integer.parseInt(args[2]));
    scriptDef.setBctype(args[3]);
    
    scriptDef.setParentDef(parentDef);
    if(args.length>4)
      scriptDef.setTag(args[4]);
    else
      scriptDef.setTag(scriptDef.getBctype()+scriptDef.getCompId());
    return scriptDef;
  }

  public static String[] elementsOfLine(String line)
  {
    return line.split("\\s+");
  }
  
  protected abstract boolean checkBeforeParse();
}
