package com.brilliance.ext.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.net.URL;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import com.brilliance.eibs.el.CommonFunctionUtils;
import com.brilliance.eibs.util.StringUtil;
import com.cfca.util.pki.api.CertUtil;
import com.cfca.util.pki.api.EncryptUtil;
import com.cfca.util.pki.api.EnvelopUtil;
import com.cfca.util.pki.api.KeyUtil;
import com.cfca.util.pki.api.SignatureUtil;
import com.cfca.util.pki.cert.X509Cert;
import com.cfca.util.pki.cipher.JCrypto;
import com.cfca.util.pki.cipher.JKey;
import com.cfca.util.pki.cipher.Session;

public class CfcaMsg {
	private Session session = null;
	private String cerPath = null;
	private String pfxPath = null;
	private String passWord = null;
	private int bufferSize=0;
	private final static int defaultBufferSize=1 * 1024 * 1024;
	
	public CfcaMsg(String passWord,String cerPath,String pfxPath) throws Exception{
		this(passWord,cerPath,pfxPath,defaultBufferSize);
	}
	
	public CfcaMsg(String passWord,String cerPath,String pfxPath,int bufferSize) throws Exception{
		URL cerPathUrl =getCommonURL(cerPath);
		URL pfxPathUrl =getCommonURL(pfxPath);
		this.cerPath = cerPathUrl==null?cerPath:cerPathUrl.getFile();
		this.pfxPath = pfxPathUrl==null?pfxPath:pfxPathUrl.getFile();	
		this.passWord= passWord;	
		this.bufferSize=bufferSize;
		JCrypto jcrypto = JCrypto.getInstance();
		jcrypto.initialize("JSOFT_LIB", null);
		this.session = jcrypto.openSession("JSOFT_LIB");
	}
	
	public URL getCommonURL(String relativePath) {
		URL url = Thread.currentThread().getContextClassLoader()
				.getResource(relativePath);
		if (url == null) {
			url = CommonFunctionUtils.class.getClassLoader().getResource(
					relativePath);
		}
		return url;
	}
	/**
	 * 对传入字符串进行CFCA加密，返回加密后字符串数据
	 * @param msgInfo 加密前字符串
	 * @return 加密后字符串
	 */
	public String envelopMsg(String msgInfo) throws Exception{
		if(StringUtil.isEmpty(msgInfo)){
			return "";
		}
		EnvelopUtil envUtil = new EnvelopUtil();
		X509Cert cert = new X509Cert(new FileInputStream(cerPath));
		envUtil.addRecipient(cert);
		envUtil.setCMSFlag();
		byte[] b64MsgEnvelop = envUtil.envelopeMessage(msgInfo.getBytes(),
				 EncryptUtil.RC4, this.session);
		return new String(b64MsgEnvelop);
	} 		
	/**
	 * 对传入字节进行CFCA加密，返回加密后字节数据
	 * @param msgInfo 加密前字节
	 * @return   加密后字节
	 */
	public byte[] envelopByte(byte[] byteInfo) throws Exception{
		if(byteInfo == null || byteInfo.length == 0 || cerPath==null || cerPath.trim().length()==0){
			return null;
		}
		EnvelopUtil envUtil = new EnvelopUtil();
		X509Cert cert = new X509Cert(new FileInputStream(cerPath));
		envUtil.addRecipient(cert);
		envUtil.setCMSFlag();
		byte[] b64MsgEnvelop = envUtil.envelopeMessage(byteInfo,
				 EncryptUtil.RC4, this.session);
		return b64MsgEnvelop;
	}
	/**
	 * 对传入字符串进行CFCA解密，返回解密后字符串数据
	 * @param msgInfo 解密前字符串
	 * @return 解密后字符串
	 */
	public String unEnvelopMsg(String msgInfo) throws Exception{
		if(StringUtil.isEmpty(msgInfo)){
			return "";
		}
		String passWord = this.passWord;
		JKey priKey = KeyUtil.getPriKey(pfxPath, passWord);
		X509Cert pfxCert = CertUtil.getCert(pfxPath, passWord);
		EnvelopUtil envUtil = new EnvelopUtil();
		X509Cert cert = new X509Cert(new FileInputStream(cerPath));
		envUtil.addRecipient(cert);
		envUtil.setCMSFlag();
		byte[] srcMsgInfo = envUtil.openEnvelopedMessage(msgInfo.getBytes(), priKey,
				pfxCert, this.session);
		return new String(srcMsgInfo);
	}	
	/**
	 * 对传入字节进行CFCA解密，返回解密后字节数据
	 * @param msgInfo 解密前字符串
	 * @return
	 */
	public byte[] unEnvelopByte(byte[] byteInfo) throws Exception{
		if(byteInfo==null || byteInfo.length ==0){
			return null;
		}
		String passWord = this.passWord;
		JKey priKey = KeyUtil.getPriKey(pfxPath, passWord);
		X509Cert pfxCert = CertUtil.getCert(pfxPath, passWord);
		EnvelopUtil envUtil = new EnvelopUtil();
		X509Cert cert = new X509Cert(new FileInputStream(cerPath));
		envUtil.addRecipient(cert);
		envUtil.setCMSFlag();
		
		byte[] srcMsgInfo = envUtil.openEnvelopedMessage(byteInfo, priKey,
				pfxCert, this.session);
		return srcMsgInfo;
	}	
	/**
     * 对文件加密
     * 原文件全路径:srcFilePath
     */
    public String envelopFile(String srcFilePath)  throws Exception{
        EnvelopUtil envUtil = new EnvelopUtil();
        X509Cert cert = new X509Cert(new FileInputStream(cerPath));
        envUtil.addRecipient(cert);
        envUtil.setCMSFlag();
        envUtil.envelopeFile(srcFilePath, srcFilePath + ".enc", bufferSize,
                             EnvelopUtil.RC4, session);
        return srcFilePath + ".enc";
    }
    /**
     *解密文件
     * 原文件全路径:envFilePath
     */
    public String openEnvelopFile(String envFilePath)  throws Exception{
        String pwd = this.passWord;
		//从私钥证书中获得数字信封接收者私钥
        JKey priKey = KeyUtil.getPriKey(pfxPath, pwd);
        //获得公钥证书
        X509Cert cert = CertUtil.getCert(pfxPath, pwd);
        EnvelopUtil envUtil = new EnvelopUtil();
        int index = envFilePath.lastIndexOf(".");
        String srcPath = envFilePath.substring(0, index);
       //解密信息
        envUtil.openEnvelopedFile(srcPath + ".dec", envFilePath, bufferSize,
                                  priKey, cert, session);
//      验证签名
        return srcPath + ".dec";
    }

    /**
     * P7文件签名
     */
    public String p7SignFile(String srcFilePath)  throws Exception{
        String pwd = this.passWord;
        //从私钥证书中获得签名者私钥
        JKey priKey = KeyUtil.getPriKey(pfxPath, pwd);
        X509Cert cert = CertUtil.getCert(pfxPath, pwd);
        SignatureUtil signUtil = new SignatureUtil();
        //对文件签名，并将签名结果输入到指定路径
        signUtil.p7SignFile(true, srcFilePath, srcFilePath + ".p7e",
                            bufferSize, SignatureUtil.SHA1_RSA, priKey, cert,
                            session);
        return srcFilePath + ".p7e";
    }
	
    /**
     * 验证签名
     */
    public String p7VerifySignFile(String signFilePath)  throws Exception{
        SignatureUtil signUtil = new SignatureUtil();
        //验证 非分离式 文件签名
        boolean verify = signUtil.p7VerifySignFile(signFilePath+".ver",
                signFilePath, bufferSize, session);
        if (verify) {
            //获得文件签名中的证书
            X509Cert[] x509Certs = signUtil.getSigerCert();
            for (int i = 0; i < x509Certs.length; i++) {
                String CertSerialNumber = x509Certs[i].getStringSerialNumber();
                X509Cert cert = CertUtil.getCert(pfxPath, this.passWord);
                if(!CertSerialNumber.toLowerCase().equals(cert.getStringSerialNumber().toLowerCase())){
                	return signFilePath+".ver";
                }
            }
        } else {
        }
        return signFilePath+".ver";
    }

    /***
	  * 字节流压缩GZip
	  * @param data
	  * @return
	  */
	public byte[] gZip(byte[] data) throws Exception{
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		GZIPOutputStream gzip = new GZIPOutputStream(bos);
		gzip.write(data);
		gzip.finish();
		gzip.close();
		byte[] b = bos.toByteArray();
		bos.close();
		return b;
	}
	 /***
	  * 字节流解压unGZip
	  * @param data
	  * @return
	  */
	public byte[] unGZip(byte[] data) throws Exception{
		ByteArrayInputStream bis = new ByteArrayInputStream(data);
		GZIPInputStream gzip = new GZIPInputStream(bis);
		byte[] buf = new byte[1024];
		int num = -1;
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		while ((num = gzip.read(buf, 0, buf.length)) != -1) {
			baos.write(buf, 0, num);
		}
		byte[] b = baos.toByteArray();
		baos.flush();
		baos.close();
		gzip.close();
		bis.close();
		return b;
	}

	 public static void main(String[] args) {
		try{
			CfcaMsg helper = new CfcaMsg("1111","F:/brilliance/workspace_myeclipse/SHDEMO/金联万家/certificate/123456789/test01.cer","F:/brilliance/workspace_myeclipse/SHDEMO/金联万家/certificate/123456789/test01.pfx");
//			 字符串加密
		String	encMsgInfo = helper.envelopMsg("wwww");
		System.out.println(encMsgInfo);
//		 字符串解密
		String decMsgInfo = helper.unEnvelopMsg(encMsgInfo);
		System.out.println("--解密字符串： " + encMsgInfo + "--结果：" + decMsgInfo + "--");
		/*	CfcaMsg helper = new CfcaMsg("111","certificate/11001604/test01.cer","certificate/11001604/test01.pfx");

//			 字符串加密
		String	encMsgInfo = helper.envelopMsg("wwww");
		System.out.println(encMsgInfo);
//		 字符串解密
		String decMsgInfo = helper.unEnvelopMsg(encMsgInfo);
		System.out.println("--解密字符串： " + encMsgInfo + "--结果：" + decMsgInfo + "--");

			*/
			/*String fileName, decMsgInfo, encMsgInfo, decFile, encFile, signFile, verSignFile;
			fileName =  System.getProperty("user.dir") + File.separator + "files" + File.separator + "公司通讯录-201512.xlsx";

			JlwjCfcaMsg helper = new JlwjCfcaMsg();

//			 字符串加密
			encMsgInfo = helper.envelopMsg(fileName);
			System.out.println("--加密字符串： " + fileName + "--结果：" + encMsgInfo + "--");
//			 字符串解密
			decMsgInfo = helper.unEnvelopMsg(encMsgInfo);
			System.out.println("--解密字符串： " + encMsgInfo + "--结果：" + decMsgInfo + "--");
//		            加密文件
			encFile = helper.envelopFile(fileName);
			System.out.println("--加密文件结果名称： " + encFile + "--");
//		            文件签名
			signFile = helper.p7SignFile(encFile);
			System.out.println("--文件签名结果： " + signFile + "--");
//		            验证文件签名
			verSignFile = helper.p7VerifySignFile(signFile);
			System.out.println("--验证文件签名结果： " + verSignFile + "--");
//			 解密文件
			decFile = helper.openEnvelopFile(verSignFile);
			System.out.println("--解密文件结果： " + decFile + "--");*/
	    } catch (Exception ex) {
	    	ex.printStackTrace();
	    }
	 }
}
