package com.brilliance.eibs.core.service.instance.plugin;

import com.brilliance.eibs.core.exception.InterfaceException;
import com.brilliance.eibs.util.StringUtil;
import org.apache.commons.io.FileExistsException;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.filefilter.RegexFileFilter;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPFile;
import org.apache.commons.net.ftp.FTPFileFilter;
import org.apache.commons.net.ftp.FTPReply;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Ftp客户端访问类 对<code>org.apache.commons.net.ftp.FTPClient</code>进行进一步封装
 * 
 * @author xiaoyuanzhen
 * 
 */
public class FtpClient {
	private FTPClient client;
	private boolean isLogined = false;
	private boolean passiveMode = false;

	// 保存远程下载文件名列表
	private List<String> filenames = new ArrayList<String>();

	/* 去除文件名为"."和".."的文件的过滤器 */
	private FTPFileFilter subFileFilter = new FTPFileFilter() {

		@Override
		public boolean accept(FTPFile f) {
			String fName = f.getName();
			if (fName.equals(".") || fName.equals(".."))
				return false;
			return true;
		}
	};

	/**
	 * FTP连接
	 * 
	 * @param host
	 *            FTP主机IP地址或名称
	 * @param port
	 *            FTP主机端口
	 */
	public FtpClient(String host, int port) {
		client = new FTPClient();
		connect(host, port);
	}

	/**
	 * FTP连接
	 * 
	 * @param host
	 *            FTP主机IP地址或名称
	 * @param port
	 *            FTP主机端口
	 * @param timeout
	 *            连接超时时间
	 */
	public FtpClient(String host, int port, int timeout) {
		client = new FTPClient();
		connect(host, port, timeout);
	}

	/**
	 * FTP连接并进行登录
	 * 
	 * @param host
	 *            FTP主机IP地址或名称
	 * @param port
	 *            FTP主机端口
	 * @param userName
	 *            登录用户名
	 * @param password
	 *            登录密码
	 */
	public FtpClient(String host, int port, String userName, String password,
			boolean passiveMode) {
		this(host, port);
		login(userName, password, passiveMode);
	}

	/**
	 * FTP连接
	 * 
	 * @param host
	 *            FTP主机IP地址或名称
	 * @param port
	 *            FTP主机端口
	 */
	private void connect(String host, int port) {
		try {
			client.connect(host, port);
		} catch (Exception e) {
			throw new InterfaceException("01207", String.format(
					"connet FTP server [%s:%d] failed.", host, port), e);
		}
	}

	/**
	 * FTP连接
	 * 
	 * @param host
	 *            FTP主机IP地址或名称
	 * @param port
	 *            FTP主机端口
	 * @param timeout
	 *            连接超时时间
	 */
	private void connect(String host, int port, int timeout) {
		client.setConnectTimeout(timeout);
		connect(host, port);
	}

	/**
	 * FTP登录
	 * 
	 * @param userName
	 *            用户名称
	 * @param password
	 *            密码
	 */
	public void login(String userName, String password, boolean passiveMode) {
		if (isLogined)
			return;
		try {
			if (!client.login(userName, password)) {
				client.disconnect();
			}
			int reply = client.getReplyCode();
			if (!FTPReply.isPositiveCompletion(reply)) {
				client.disconnect();
			} else {
				isLogined = true;
			}
			if (passiveMode)
				client.enterLocalPassiveMode();
			// 设置文件传输模式
			client.setFileType(FTPClient.BINARY_FILE_TYPE);
		} catch (Exception e) {
			throw new InterfaceException("01207",
					String.format("ftp login failed."), e);
		}
		if (!isLogined)
			throw new InterfaceException("01207",
					String.format("ftp login failed."));
	}

	public void setBinaryMode(boolean b) throws IOException {
		client.setFileType(b ? FTPClient.BINARY_FILE_TYPE
				: FTPClient.ASCII_FILE_TYPE);
	}

	/**
	 * 切换工作目录
	 * 
	 * @param dirName
	 * @exception 登录失败报错
	 */
	private void cd(String dirName) {
		try {
			if (!client.changeWorkingDirectory(dirName)) {
				throw new InterfaceException("01211", "invalid remote path:"
						+ dirName);
			}
		} catch (IOException e) {
			throw new InterfaceException("01211", "invalid remote path:"
					+ dirName, e);
		}
	}

	/**
	 * 上传文件
	 * 
	 * @param localPath
	 *            上传的本地路径
	 * @param remotePath
	 *            上传的远程路径
	 * @param filenameRegex
	 *            上传目录过滤(第一层子目录，无递归)
	 * @param backupDirName
	 *            上传备份目录
	 * @exception 非法的本地路径
	 */
	public void upload(String localPath, String remotePath,
			String filenameRegex, String backupDirName) {
		File file = new File(localPath);
		File[] subFiles = null;
		if (file.exists()) {
			try {
				client.makeDirectory(remotePath);
			} catch (IOException e) {
				throw new InterfaceException("01213", "make dirs error.", e);
			}
			cd(remotePath);
			if (file.isDirectory()) {
				if (!StringUtil.isEmpty(filenameRegex))
					subFiles = file
							.listFiles((FilenameFilter) new RegexFileFilter(
									filenameRegex));
				else
					subFiles = file.listFiles();
				for (int i = 0; i < subFiles.length; i++) {
					upload(subFiles[i]);
				}
			} else {
				subFiles = new File[] { file };
				upload(file);
			}
		} else {
			throw new InterfaceException("01215", "invalid local path.");
		}
		if (!StringUtil.isEmpty(backupDirName)) {
			File backupDir = new File(backupDirName);
			try {
				for (int i = 0; i < subFiles.length; i++) {
					File srcFile = subFiles[i];
					try {
						FileUtils.moveFileToDirectory(srcFile, backupDir, true);
					} catch (FileExistsException e) {
						File destFile = new File(backupDir, srcFile.getName());
						if (destFile.delete())
							FileUtils.moveFile(srcFile, destFile);
					}
				}
			} catch (IOException e) {
				throw new InterfaceException("01218", "upload backup error.", e);
			}
		}
	}

	/**
	 * 上传文件(夹)
	 * 
	 * @param file
	 * @exception 上传失败
	 */
	private void upload(File file) {
		try {
			uploadThrow(file);
		} catch (IOException e) {
			throw new InterfaceException("01210", "ftp upload file error: "
					+ file, e);
		}
	}

	/**
	 * 上传文件(夹)
	 * 
	 * @param file
	 * @exception 上传失败
	 */
	private void uploadThrow(File file) throws IOException {
		if (file.isDirectory()) {
			boolean flg = true;
			String dirName = file.getName();
			client.makeDirectory(dirName);
			client.changeWorkingDirectory(dirName);
			File[] subFiles = file.listFiles();
			for (int i = 0; i < subFiles.length; i++) {
				upload(subFiles[i]);
				if (subFiles[i].isDirectory()) {
					client.changeToParentDirectory();
					flg = false;
				}
			}
			// 保证至少能执行一次
			if (flg)
				client.changeToParentDirectory();
		} else {
			InputStream stream = new FileInputStream(file);
			client.storeFile(file.getName(), stream);
			IOUtils.closeQuietly(stream);
		}
	}

	/**
	 * FTP下载文件(夹)
	 * 
	 * @param rmtPath
	 *            远程目录
	 * @param localPath
	 *            本地目录
	 * @param filenameRegex
	 *            下载过滤正则式
	 * @param backupDirName
	 *            下载后备份的目录
	 */
	public void download(String ctlPath, final String ctlFilenameregex,
			String rmtPath, String localPath, final String filenameRegex,
			String backupDirName, boolean deleteFlag) {
		try {
			FTPFile[] ctlFiles = null;
			if (!StringUtil.isEmpty(ctlPath)) {
				cd(ctlPath);
				ctlFiles = client.listFiles(ctlPath, new FTPFileFilter() {
					@Override
					public boolean accept(FTPFile f) {
						if (f.isFile()) {
							if (f.getName().matches(ctlFilenameregex))
								return true;
							return false;
						}
						return false;
					}
				});
			}

			if (isRemoteDir(rmtPath)) {
				cd(rmtPath);
				FTPFile[] subFiles = null;
				if (!StringUtil.isEmpty(filenameRegex)) {
					subFiles = client.listFiles(rmtPath, new FTPFileFilter() {
						@Override
						public boolean accept(FTPFile f) {
							if (f.getName().matches(filenameRegex))
								return true;
							return false;
						}
					});
				} else
					subFiles = client.listFiles(rmtPath, subFileFilter);
				for (int i = 0; i < subFiles.length; i++) {
					if (!containCtlFile(ctlFiles, subFiles[i])) {
						download(subFiles[i], localPath);

					}
					filenames.add(subFiles[i].getName());
				}
				if (!StringUtil.isEmpty(backupDirName)) {
					for (int i = 0; i < subFiles.length; i++) {
						if (!containCtlFile(ctlFiles, subFiles[i])) {
							String subFileName = subFiles[i].getName();
							String from = rmtPath + "/" + subFileName;
							String to = backupDirName + "/" + subFileName;
							rename(from, to);
						}
					}
				}
				if (StringUtil.isEmpty(backupDirName) && deleteFlag) {
					for (int i = 0; i < subFiles.length; i++) {
						if (!containCtlFile(ctlFiles, subFiles[i])) {
							String filepath = pwd() + "/"
									+ subFiles[i].getName();
							client.deleteFile(filepath);
						}
					}
				}
			} else if (isRemoteFile(rmtPath)) {
				FTPFile file = getRemoteFile(rmtPath);
				File rmtFile = new File(rmtPath);
				String pName = getParentPath(rmtFile);
				String pwd = pwd();
				cd(pName);
				if (file != null) {
					if (!containCtlFile(ctlFiles, file)) {
						download(file, localPath);
					}
					filenames.add(file.getName());
				}
				if (!StringUtil.isEmpty(backupDirName)) {
					if (!containCtlFile(ctlFiles, file)) {
						String subFileName = file.getName();
						String from = pwd() + "/" + subFileName;
						String to = backupDirName + "/" + subFileName;
						rename(from, to);
					}
				}
				if (StringUtil.isEmpty(backupDirName) && deleteFlag) {
					if (!containCtlFile(ctlFiles, file)) {
						String filepath = pwd() + "/" + file.getName();
						client.deleteFile(filepath);
					}
				}
				cd(pwd);
			} else {
				throw new InterfaceException("01211", "invalid remote path:"
						+ rmtPath);
			}
		} catch (IOException e) {
			throw new InterfaceException("01212", "FTP文件下载异常", e);
		}
	}

	private boolean containCtlFile(FTPFile[] ctlFiles, FTPFile f) {
		if (ctlFiles == null)
			return false;
		if (f == null)
			return false;
		for (int j = 0; j < ctlFiles.length; j++) {
			FTPFile currentFile = ctlFiles[j];
			if (currentFile != null) {
				String currenFileName = currentFile.getName();
				String downFileName = f.getName();

				String cName = currenFileName.substring(0, len(currenFileName));
				String fName = downFileName.substring(0, len(downFileName));
				if (cName.equals(fName))
					return true;
			}

		}
		return false;
	}

	public int len(String fileName) {
		int lastIndex = fileName.lastIndexOf(".");
		return lastIndex == -1 ? fileName.length() : lastIndex;
	}

	/**
	 * 获取当前工作目录
	 * 
	 * @return
	 */
	private String pwd() {
		try {
			return client.printWorkingDirectory();
		} catch (IOException e) {
			throw new InterfaceException("01217", "get pwd error.", e);
		}
	}

	/**
	 * 根据指定的远程路径获取FTPFile对象
	 * 
	 * @param rmtPath
	 *            远程路径
	 * @return
	 */
	private FTPFile getRemoteFile(String rmtPath) {
		File file = new File(rmtPath);
		String pName = getParentPath(file);
		final String fileName = file.getName();
		String pwd = pwd();
		cd(pName);
		FTPFile[] subFiles = null;
		try {
			subFiles = client.listFiles(pwd(), new FTPFileFilter() {

				@Override
				public boolean accept(FTPFile f) {
					if (f.getName().equals(fileName))
						return true;
					return false;
				}
			});
		} catch (IOException e) {
		}
		cd(pwd);
		if (subFiles != null && subFiles.length > 0)
			return subFiles[0];
		return null;
	}

	/**
	 * FTP下载文件
	 * 
	 * @param file
	 * @param localPath
	 * @exception 下载失败
	 */
	private void download(FTPFile file, String localPath) {
		try {
			downloadThrow(file, localPath);
		} catch (IOException e) {
			throw new InterfaceException("01212", "ftp download file error:"
					+ file.getName(), e);
		}
	}

	/**
	 * FTP下载文件
	 * 
	 * @param file
	 * @param localPath
	 * @throws 下载失败
	 */
	private void downloadThrow(FTPFile file, String localPath)
			throws IOException {
		String fName = null;
		fName = file.getName();
		File localFile = new File(localPath + "/" + fName);
		File localParentDir = localFile.getParentFile();
		if (!localParentDir.exists())
			localParentDir.mkdirs();
		if (file.isDirectory()) {
			client.changeWorkingDirectory(fName);
			FTPFile[] subFiles = client.listFiles(
					client.printWorkingDirectory(), subFileFilter);
			for (int i = 0; i < subFiles.length; i++) {
				download(subFiles[i], localPath + "/" + fName);
			}
			client.changeToParentDirectory();
		} else {
			OutputStream os = new FileOutputStream(localFile);
			client.retrieveFile(fName, os);
			IOUtils.closeQuietly(os);
		}
	}

	/**
	 * 文件(夹)重定向
	 * 
	 * @param from
	 * @param to
	 * @exception 非法的远程路径
	 *                、重命名失败
	 */
	public void rename(String from, String to) {
		try {
			File file = new File(to);
			if (remoteFileExists(from)) {
				String pName = getParentPath(file);
				if (pName != null)
					mkdirs(pName);
				if (remoteFileExists(to)) {
					client.deleteFile(to);
				}
				if (!client.rename(from, to))
					throw new InterfaceException("01214", "rename file from ["
							+ from + "] to " + to + " failed.");
				/*
				 * try { if (!client.rename(from, to)); } catch (Exception e) {
				 * e.printStackTrace(); }
				 */
			} else
				throw new InterfaceException("01211", "invalid remote path:"
						+ from);
		} catch (Exception e) {
			throw new InterfaceException("01214", "rename file from [" + from
					+ "] to " + to + " failed.", e);
		}
	}

	/**
	 * 递归创建远程目录
	 * 
	 * @param dirName
	 * @exception 创建目录失败
	 */
	private void mkdirs(String dirName) {
		try {
			String pwd = pwd();
			mkdirsThrow(dirName);
			cd(pwd);
		} catch (IOException e) {
			throw new InterfaceException("01213", "make dirs error.", e);
		}
	}

	/**
	 * 指定路径是否是远程目录
	 * 
	 * @param dirName
	 * @return
	 */
	private boolean isRemoteDir(String dirName) {
		boolean isRemoteDir = false;
		try {
			String pwd = client.printWorkingDirectory();
			if (client.changeWorkingDirectory(dirName)) {
				isRemoteDir = true;
				client.changeWorkingDirectory(pwd);
			}
		} catch (IOException e) {
		}
		return isRemoteDir;
	}

	/**
	 * 指定路径是否是远程文件
	 * 
	 * @param fileName
	 * @return
	 */
	private boolean isRemoteFile(String fileName) {
		boolean isRemoteFile = false;
		try {
			String pwd = client.printWorkingDirectory();
			File file = new File(fileName);
			String curName = file.getName();
			String pName = getParentPath(file);
			boolean b1 = client.changeWorkingDirectory(pName);
			String[] files = client.listNames();
			List<String> names = Arrays.asList(files);
			boolean b2 = names.contains(curName);
			if (b1 && b2) {
				isRemoteFile = true;
			}
			client.changeWorkingDirectory(pwd);
		} catch (IOException e) {
		}
		return isRemoteFile;
	}

	/**
	 * 远程文件(夹)是否存在
	 * 
	 * @param name
	 * @return
	 */
	private boolean remoteFileExists(String name) {
		return isRemoteDir(name) || isRemoteFile(name);
	}

	/**
	 * 递归创建远程目录
	 * 
	 * @param dirName
	 * @throws IOException
	 */
	private void mkdirsThrow(String dirName) throws IOException {
		File file = new File(dirName);
		String curDirName = file.getName();
		String pDirName = getParentPath(file);
		if (pDirName != null) {
			mkdirsThrow(pDirName);
			client.changeWorkingDirectory(pDirName);
		}
		client.makeDirectory(curDirName);
	}

	/**
	 * 关闭FTP连接
	 */
	public void disConnect() {
		try {
			if (isLogined) {
				client.logout();
				isLogined = false;
			}
			client.disconnect();
		} catch (IOException e) {
			throw new InterfaceException("01216", "FTP disconnect error.", e);
		}
	}

	private static String getParentPath(File f) {
		return FilenameUtils.separatorsToUnix(f.getParent());
	}

	public List<String> getFilenames() {
		return filenames;
	}

}
