package com.brilliance.eibs.core.service.instance.plugin;

import com.brilliance.eibs.core.frame.ISocketProxy;
import com.brilliance.eibs.core.service.Context;
import com.brilliance.eibs.util.LogUtil;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;

import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.SocketAddress;

/**
 * Socket的代理类，用来包装Socket
 *
 * @author hujun
 */
public class SocketProxy implements ISocketProxy {
    private Logger log;
    private String encoding;
    private Socket socket;
    private Context context;

    public SocketProxy(SocketAddress addrress, int connectTimeout, int receiveTimeout, String encoding, Context context) throws IOException {
        this.log = LogUtil.getLogger(context);
        this.context = context;
        this.socket = new Socket();
        this.encoding = encoding;

        socket.setSoTimeout(receiveTimeout);
        socket.connect(addrress, connectTimeout);

        log.debug("Create Socket agent successfully.");
    }

    public SocketProxy(Socket socket, String encoding, Context context) {
        this.log = LogUtil.getLogger(context);
        this.context = context;
        this.socket = socket;
        this.encoding = encoding;
        log.debug("Create Socket agent successfully.");
    }

    @Override
    public void close() {
        IOUtils.closeQuietly(socket);
    }

    public String getEncoding() {
        return encoding;
    }

    @Override
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }

    /**
     * 获取socket流中的len长度的字节对应的字符串
     *
     * @param len
     * @return
     * @throws IOException
     */
    @Override
    public String getDataStream(int len) throws IOException {
        String s = new String(getDataStreamBytes(len), encoding);
        log.debug("Read the content of input stream is " + s);
        return s;
    }

    @Override
    public byte[] getDataStreamBytes(int len) throws IOException {
        log.debug("Read [" + len + " bytes] from input stream");
        byte[] buf = new byte[len];
        if (socket.isClosed()) {
            log.debug("socket is closed before read.");
        }
        IOUtils.readFully(socket.getInputStream(), buf);
        // log.info("[" + encoding + "]socket getdatabytes=" + new String(buf,
        // encoding));
        return buf;
    }

    /**
     * 循环读取 获取socket流中的报文内容
     *
     * @return
     * @throws IOException
     */
    @Override
    public String getStrBySocketStream() throws IOException {
        String s = null;
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        int len = 0;
        byte buffer[] = new byte[100];
        while ((len = socket.getInputStream().read(buffer)) != -1) {
            bos.write(buffer, 0, len);
            byte buf[] = bos.toByteArray();
            s = new String(buf, encoding);
            int index = s.indexOf(">");
            if (index != -1) {
                String ss = s.substring(index + 1);
                if (ss.startsWith("<") && ss.endsWith("/>")) {
                    return s;
                }
                int index2 = ss.indexOf("<");
                int index3 = ss.indexOf(">");
                if (index2 != -1 && index3 != -1) {
                    String xmlroot = ss.substring(index2 + 1, index3);
                    if (ss.contains("</" + xmlroot + ">")) {
                        return s;
                    }
                }
            }
        }
        return null;
    }

    /**
     * 输出字符串到socket
     *
     * @param message
     * @throws IOException
     */
    public void send(Object message) throws IOException {
        if (message == null) {
            log.debug("No message to send.");
            return;
        }
        byte[] msgbytes = null;
        if (message instanceof String) {
            msgbytes = ((String) message).getBytes(encoding);
        } else if (message instanceof byte[]) {
            msgbytes = (byte[]) message;
        }
        log.trace("Write [" + msgbytes.length + " bytes] to output stream, content is " + message);
        if (isSocketValid(socket)) {
            log.debug("Start sending message.");
            IOUtils.write(msgbytes, socket.getOutputStream());
        }
    }

    /**
     * 将一个数字写入到4个字节当中(高位字节)
     *
     * @param length
     * @throws IOException
     */
    public void writeInt(int length) throws IOException {
        log.debug("Write int = " + length);
        DataOutputStream dos = new DataOutputStream(socket.getOutputStream());
        dos.writeInt(length);
    }

    /**
     * 读取输入流中的4个字节作为一个数字返回
     *
     * @return
     * @throws IOException
     */
    public int readInt() throws IOException {
        DataInputStream dis = new DataInputStream(socket.getInputStream());
        int len = dis.readInt();
        log.debug("read int = " + len);
        return len;
    }

    /**
     * 跳过输入流的len字节
     *
     * @param len
     * @throws IOException
     */
    public void skip(long len) throws IOException {
        log.debug("Skip [" + len + " bytes] from input stream.");
        IOUtils.skipFully(socket.getInputStream(), len);
    }

    /**
     * 判断socket是否有效，除了判断socket是否被close还需要判断
     *
     * @param socket
     * @return
     */
    private boolean isSocketValid(Socket socket) {
        boolean closed = socket.isClosed() || !socket.isConnected();
        if (!closed) {
            try {
                // 发送远程
                // socket.sendUrgentData(0xff);
                return true;
            } catch (Exception e) {
                return false;
            }
        }
        return false;
    }

    @Override
    public boolean isConnected() {
        return !socket.isClosed() && socket.isConnected();
    }

    @Override
    public Object[] readDataByBuffer() throws IOException {
        Object[] obj = new Object[2];
        byte[] buffer = new byte[1024];
        int len = -1;
        byte[] temp = null;
        len = socket.getInputStream().read(buffer);
        if (len != -1) {
            temp = new byte[len];
            System.arraycopy(buffer, 0, temp, 0, len);
        }
        obj[0] = len;
        obj[1] = temp;

        return obj;

    }

    public Socket getSocket() {
        return socket;
    }

    @Override
    public Object getObject() throws IOException {
        // TODO Auto-generated method stub
        return null;
    }
}
