package com.brilliance.eibs.factory.parser;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.io.SAXValidator;
import org.dom4j.util.XMLErrorHandler;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.util.List;

/**
 * Xml校验器
 * 
 * @author xiaoyuanzhen
 *
 */
public class XmlValidator {
	private String strErrors; // 错误描述

	/**
	 * 校验Document文档对象
	 * 
	 * @param doc
	 * @return
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public boolean validate(Document doc) throws ParserConfigurationException, SAXException {
		SAXParser parser = getSAXParser();
		return validate(doc, parser);
	}

	/**
	 * 使用指定路径的xsd文件校验Document文档对象
	 * 
	 * @param doc
	 * @param xsdPath
	 * @return
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public boolean validate(Document doc, String xsdPath) throws ParserConfigurationException, SAXException {
		SAXParser parser = getSAXParser(xsdPath);
		return validate(doc, parser);
	}

	private boolean validate(Document doc, SAXParser parser) throws ParserConfigurationException, SAXException {
		XMLErrorHandler handler = new XMLErrorHandler();
		SAXValidator validator = new SAXValidator(parser.getXMLReader());
		validator.setErrorHandler(handler);
		validator.validate(doc);
		if (handler.getErrors().hasContent()) {
			Element errorsNode = handler.getErrors();
			List<Element> errorNodes = errorsNode.elements();
			StringBuffer errorBuffer = new StringBuffer();
			String lineSep = System.getProperty("line.separator");
			errorBuffer.append("xml validate error." + lineSep);
			int errorSize = errorNodes.size();
			for (int i = 0; i < errorSize; i++) {
				Element errorNode = errorNodes.get(i);
				String column = errorNode.attributeValue("column");
				String line = errorNode.attributeValue("line");
				String systemID = errorNode.attributeValue("systemID");
				String errorText = errorNode.getText();
				errorBuffer.append("Error: [" + systemID + "](line:" + line + ",column:" + column + ") : " + lineSep + errorText);
				if (i != errorSize - 1)
					errorBuffer.append(lineSep);
			}
			strErrors = errorBuffer.toString();
			return false;
		}
		return true;
	}

	/**
	 * 使用指定路径的xsd文件校验xml字符串格式
	 * 
	 * @param xmlstr
	 * @param xsdPath
	 * @return
	 * @throws DocumentException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public boolean validate(String xmlstr, String xsdPath) throws DocumentException, ParserConfigurationException, SAXException {
		Document doc = DocumentHelper.parseText(xmlstr);
		return validate(doc, xsdPath);
	}

	/**
	 * 获取默认SAXParser对象
	 * 
	 * @return
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	private SAXParser getSAXParser() throws ParserConfigurationException, SAXException {
		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setValidating(true);
		factory.setNamespaceAware(true);
		SAXParser parser = factory.newSAXParser();
		parser.setProperty("http://java.sun.com/xml/jaxp/properties/schemaLanguage", "http://www.w3.org/2001/XMLSchema");
		return parser;
	}

	/**
	 * 获取含xsd路径信息的SAXParser
	 * 
	 * @param xsdPath
	 * @return
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	private SAXParser getSAXParser(String xsdPath) throws ParserConfigurationException, SAXException {
		SAXParser parser = getSAXParser();
		parser.setProperty("http://java.sun.com/xml/jaxp/properties/schemaSource", xsdPath);
		return parser;
	}

	public String getStrErrors() {
		return strErrors;
	}

}
