package com.brilliance.eibs.quartz;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.text.ParseException;
import java.util.*;

public final class CronExpression
    implements Serializable, Cloneable {
  private static final long serialVersionUID = 12423409423L;
  public static final int MAX_YEAR = Calendar.getInstance().get(1) + 100;
  protected static final int SECOND = 0;
  protected static final int MINUTE = 1;
  protected static final int HOUR = 2;
  protected static final int DAY_OF_MONTH = 3;
  protected static final int MONTH = 4;
  protected static final int DAY_OF_WEEK = 5;
  protected static final int YEAR = 6;
  protected static final int ALL_SPEC_INT = 99;
  protected static final int NO_SPEC_INT = 98;
  protected static final Integer ALL_SPEC = Integer.valueOf(99);
  protected static final Integer NO_SPEC = Integer.valueOf(98);
  protected static final Map<String, Integer> monthMap = new HashMap(20);
  protected static final Map<String, Integer> dayMap = new HashMap(60);
  private final String cronExpression;
  protected transient TreeSet<Integer> seconds;
  protected transient TreeSet<Integer> minutes;
  protected transient TreeSet<Integer> hours;
  protected transient TreeSet<Integer> daysOfMonth;
  protected transient TreeSet<Integer> months;
  protected transient TreeSet<Integer> daysOfWeek;
  protected transient TreeSet<Integer> years;
  protected transient boolean lastdayOfWeek = false;
  protected transient int nthdayOfWeek = 0;
  protected transient boolean lastdayOfMonth = false;
  protected transient boolean nearestWeekday = false;
  protected transient boolean expressionParsed = false;
  private TimeZone timeZone = null;

  public CronExpression(String cronExpression)
      throws ParseException {
    if (cronExpression == null) {
      throw new IllegalArgumentException("cronExpression cannot be null");
    }
    this.cronExpression = cronExpression.toUpperCase(Locale.US);
    buildExpression(this.cronExpression);
  }

  public static boolean isValidExpression(String cronExpression) {
    try {
      new CronExpression(cronExpression);
    } catch (ParseException pe) {
      return false;
    }

    return true;
  }

  public static void validateExpression(String cronExpression) throws ParseException {
    new CronExpression(cronExpression);
  }

  public boolean isSatisfiedBy(Date date) {
    Calendar testDateCal = Calendar.getInstance(getTimeZone());
    testDateCal.setTime(date);
    testDateCal.set(14, 0);
    Date originalDate = testDateCal.getTime();
    testDateCal.add(13, -1);
    Date timeAfter = getTimeAfter(testDateCal.getTime());
    return (timeAfter != null) && (timeAfter.equals(originalDate));
  }

  public Date getNextValidTimeAfter(Date date) {
    return getTimeAfter(date);
  }

  public Date getNextInvalidTimeAfter(Date date) {
    long difference = 1000L;
    Calendar adjustCal = Calendar.getInstance(getTimeZone());
    adjustCal.setTime(date);
    adjustCal.set(14, 0);
    Date lastDate = adjustCal.getTime();
    while (difference == 1000L) {
      Date newDate = getTimeAfter(lastDate);
      if (newDate == null) {
        break;
      }
      difference = newDate.getTime() - lastDate.getTime();

      if (difference == 1000L) {
        lastDate = newDate;
      }
    }
    return new Date(lastDate.getTime() + 1000L);
  }

  public TimeZone getTimeZone() {
    if (this.timeZone == null) {
      this.timeZone = TimeZone.getDefault();
    }

    return this.timeZone;
  }

  public void setTimeZone(TimeZone timeZone) {
    this.timeZone = timeZone;
  }

  public String toString() {
    return this.cronExpression;
  }

  protected void buildExpression(String expression)
      throws ParseException {
    this.expressionParsed = true;
    try {
      if (this.seconds == null) {
        this.seconds = new TreeSet();
      }
      if (this.minutes == null) {
        this.minutes = new TreeSet();
      }
      if (this.hours == null) {
        this.hours = new TreeSet();
      }
      if (this.daysOfMonth == null) {
        this.daysOfMonth = new TreeSet();
      }
      if (this.months == null) {
        this.months = new TreeSet();
      }
      if (this.daysOfWeek == null) {
        this.daysOfWeek = new TreeSet();
      }
      if (this.years == null) {
        this.years = new TreeSet();
      }
      int exprOn = 0;
      for (StringTokenizer exprsTok = new StringTokenizer(expression, " \t", false);
          exprsTok.hasMoreTokens() && exprOn <= 6; exprOn++) {
        String expr = exprsTok.nextToken().trim();
        StringTokenizer vTok = new StringTokenizer(expr, ",");
        while (vTok.hasMoreTokens()) {
          String v = vTok.nextToken();
          storeExpressionVals(0, v, exprOn);
        }
      }
      if (exprOn <= 5) {
        throw new ParseException("Unexpected end of expression.", expression.length());
      }
      if (exprOn <= 6) {
        storeExpressionVals(0, "*", 6);
      }
    } catch (ParseException pe) {
      throw pe;
    } catch (Exception e) {
      throw new ParseException(
          new StringBuilder().append("Illegal cron expression format (").append(e.toString())
              .append(")").toString(), 0);
    }
  }

  protected int storeExpressionVals(int pos, String s, int type)
      throws ParseException {
    int incr = 0;
    int i = skipWhiteSpace(pos, s);
    if (i >= s.length()) {
      return i;
    }
    char c = s.charAt(i);
    if (c >= 'A' && c <= 'Z' && !s.equals("L") && !s.equals("LW")) {
      String sub = s.substring(i, i + 3);
      int sval = -1;
      int eval = -1;
      if (type == 4) {
        sval = getMonthNumber(sub) + 1;
        if (sval < 0) {
          throw new ParseException(
              new StringBuilder().append("Invalid Month value: '").append(sub).append("'")
                  .toString(), i);
        }
        if (s.length() > i + 3) {
          c = s.charAt(i + 3);
          if (c == '-') {
            i += 4;
            sub = s.substring(i, i + 3);
            eval = getMonthNumber(sub) + 1;
            if (eval < 0) {
              throw new ParseException(
                  new StringBuilder().append("Invalid Month value: '").append(sub).append("'")
                      .toString(), i);
            }
          }
        }
      } else if (type == 5) {
        sval = getDayOfWeekNumber(sub);
        if (sval < 0) {
          throw new ParseException(
              new StringBuilder().append("Invalid Day-of-Week value: '").append(sub).append("'")
                  .toString(), i);
        }
        if (s.length() > i + 3) {
          c = s.charAt(i + 3);
          if (c == '-') {
            i += 4;
            sub = s.substring(i, i + 3);
            eval = getDayOfWeekNumber(sub);
            if (eval < 0) {
              throw new ParseException(
                  new StringBuilder().append("Invalid Day-of-Week value: '").append(sub).append("'")
                      .toString(), i);
            }
            if (sval > eval) {
              throw new ParseException(
                  new StringBuilder().append("Invalid Day-of-Week sequence: ").append(sval)
                      .append(" > ").append(eval).toString(), i);
            }
          } else if (c == '#') {
            try {
              i += 4;
              this.nthdayOfWeek = Integer.parseInt(s.substring(i));
              if (this.nthdayOfWeek < 1 || this.nthdayOfWeek > 5) {
                throw new Exception();
              }
            } catch (Exception e) {
              throw new ParseException("A numeric value between 1 and 5 must follow the '#' option",
                  i);
            }
          } else if (c == 'L') {
            this.lastdayOfWeek = true;
            i++;
          }
        }
      } else {
        throw new ParseException(
            new StringBuilder().append("Illegal characters for this position: '").append(sub)
                .append("'").toString(), i);
      }
      if (eval != -1) {
        incr = 1;
      }
      addToSet(sval, eval, incr, type);
      return i + 3;
    }
    if (c == '?') {
      if (++i + 1 < s.length() && s.charAt(i) != ' ' && s.charAt(i + 1) != '\t') {
        throw new ParseException(
            new StringBuilder().append("Illegal character after '?': ").append(s.charAt(i))
                .toString(), i);
      }
      if (type != 5 && type != 3) {
        throw new ParseException("'?' can only be specified for Day-of-Month or Day-of-Week.", i);
      }
      if (type == 5 && !lastdayOfMonth) {
        int val = ((Integer) daysOfMonth.last()).intValue();
        if (val == 98) {
          throw new ParseException("'?' can only be specfied for Day-of-Month -OR- Day-of-Week.",
              i);
        }
      }
      addToSet(98, -1, 0, type);
      return i;
    }
    if (c == '*' || c == '/') {
      if ((c == '*') && (i + 1 >= s.length())) {
        addToSet(99, -1, incr, type);
        return i + 1;
      }
      if ((c == '/') && ((i + 1 >= s.length()) || (s.charAt(i + 1) == ' ') || (s.charAt(i + 1)
          == '\t'))) {
        throw new ParseException("'/' must be followed by an integer.", i);
      }
      if (c == '*') {
        i++;
      }
      c = s.charAt(i);
      if (c == '/') {
        if (++i >= s.length()) {
          throw new ParseException("Unexpected end of string.", i);
        }
        incr = getNumericValue(s, i);
        i++;
        if (incr > 10) {
          i++;
        }
        checkIncrementRange(incr, type, i);
      } else {
        incr = 1;
      }
      addToSet(99, -1, incr, type);
      return i;
    }
    if (c == 'L') {
      i++;
      if (type == 3) {
        this.lastdayOfMonth = true;
      }
      if (type == 5) {
        addToSet(7, 7, 0, type);
      }
      if (type == 3 && s.length() > i) {
        c = s.charAt(i);
        if (c == 'W') {
          this.nearestWeekday = true;
          i++;
        }
      }
      return i;
    }
    if (c >= '0' && c <= '9') {
      int val = Integer.parseInt(String.valueOf(c));
      if (++i >= s.length()) {
        addToSet(val, -1, -1, type);
      } else {
        c = s.charAt(i);
        if (c >= '0' && c <= '9') {
          ValueSet vs = getValue(val, s, i);
          val = vs.value;
          i = vs.pos;
        }
        i = checkNext(i, s, val, type);
        return i;
      }
    } else {
      throw new ParseException("Unexpected character: " + c, i);
    }
    return i;
  }

  private void checkIncrementRange(int incr, int type, int idxPos) throws ParseException {
    if ((incr > 59) && ((type == 0) || (type == 1))) {
      throw new ParseException(
          new StringBuilder().append("Increment > 60 : ").append(incr).toString(), idxPos);
    }
    if ((incr > 23) && (type == 2)) {
      throw new ParseException(
          new StringBuilder().append("Increment > 24 : ").append(incr).toString(), idxPos);
    }
    if ((incr > 31) && (type == 3)) {
      throw new ParseException(
          new StringBuilder().append("Increment > 31 : ").append(incr).toString(), idxPos);
    }
    if ((incr > 7) && (type == 5)) {
      throw new ParseException(
          new StringBuilder().append("Increment > 7 : ").append(incr).toString(), idxPos);
    }
    if ((incr > 12) && (type == 4)) {
      throw new ParseException(
          new StringBuilder().append("Increment > 12 : ").append(incr).toString(), idxPos);
    }
  }

  protected int checkNext(int pos, String s, int val, int type)
      throws ParseException {
    int end = -1;
    int i = pos;
    if (i >= s.length()) {
      addToSet(val, end, -1, type);
      return i;
    }
    char c = s.charAt(pos);
    if (c == 'L') {
      if (type == 5) {
        if ((val < 1) || (val > 7)) {
          throw new ParseException("Day-of-Week values must be between 1 and 7", -1);
        }
        this.lastdayOfWeek = true;
      } else {
        throw new ParseException(
            new StringBuilder().append("'L' option is not valid here. (pos=").append(i).append(")")
                .toString(), i);
      }
      TreeSet set = getSet(type);
      set.add(Integer.valueOf(val));
      return ++i;
    }
    if (c == 'W') {
      if (type == 3) {
        nearestWeekday = true;
      } else {
        throw new ParseException(
            new StringBuilder().append("'W' option is not valid here. (pos=").append(i).append(")")
                .toString(), i);
      }
      if (val > 31) {
        throw new ParseException(
            "The 'W' option does not make sense with values larger than 31 (max number of days in a month)",
            i);
      }
      TreeSet set = getSet(type);
      set.add(Integer.valueOf(val));
      return ++i;
    }
    if (c == '#') {
      if (type != 5) {
        throw new ParseException(
            new StringBuilder().append("'#' option is not valid here. (pos=").append(i).append(")")
                .toString(), i);
      }
      i++;
      try {
        this.nthdayOfWeek = Integer.parseInt(s.substring(i));
        if (this.nthdayOfWeek < 1 || this.nthdayOfWeek > 5) {
          throw new Exception();
        }
      } catch (Exception e) {
        throw new ParseException("A numeric value between 1 and 5 must follow the '#' option", i);
      }
      TreeSet set = getSet(type);
      set.add(Integer.valueOf(val));
      return ++i;
    }
    if (c == '-') {
      i++;
      c = s.charAt(i);
      int v = Integer.parseInt(String.valueOf(c));
      end = v;
      if (++i >= s.length()) {
        addToSet(val, end, 1, type);
        return i;
      }
      c = s.charAt(i);
      if ((c >= '0') && (c <= '9')) {
        ValueSet vs = getValue(v, s, i);
        end = vs.value;
        i = vs.pos;
      }
      if ((i < s.length()) && ((c = s.charAt(i)) == '/')) {
        i++;
        c = s.charAt(i);
        int v2 = Integer.parseInt(String.valueOf(c));
        if (++i >= s.length()) {
          addToSet(val, end, v2, type);
          return i;
        }
        c = s.charAt(i);
        if ((c >= '0') && (c <= '9')) {
          ValueSet vs = getValue(v2, s, i);
          int v3 = vs.value;
          addToSet(val, end, v3, type);
          i = vs.pos;
          return i;
        }
        addToSet(val, end, v2, type);
        return i;
      }
      addToSet(val, end, 1, type);
      return i;
    }
    if (c == '/') {
      if ((i + 1 >= s.length()) || (s.charAt(i + 1) == ' ') || (s.charAt(i + 1) == '\t')) {
        throw new ParseException("'/' must be followed by an integer.", i);
      }
      i++;
      c = s.charAt(i);
      int v2 = Integer.parseInt(String.valueOf(c));
      if (++i >= s.length()) {
        checkIncrementRange(v2, type, i);
        addToSet(val, end, v2, type);
        return i;
      }
      c = s.charAt(i);
      if ((c >= '0') && (c <= '9')) {
        ValueSet vs = getValue(v2, s, i);
        int v3 = vs.value;
        checkIncrementRange(v3, type, i);
        addToSet(val, end, v3, type);
        i = vs.pos;
        return i;
      }
      throw new ParseException(
          new StringBuilder().append("Unexpected character '").append(c).append("' after '/'")
              .toString(), i);
    }
    addToSet(val, end, 0, type);
    return ++i;
  }

  public String getCronExpression() {
    return this.cronExpression;
  }

  public String getExpressionSummary() {
    StringBuilder buf = new StringBuilder();
    buf.append("seconds: ");
    buf.append(getExpressionSetSummary(this.seconds));
    buf.append("\n");
    buf.append("minutes: ");
    buf.append(getExpressionSetSummary(this.minutes));
    buf.append("\n");
    buf.append("hours: ");
    buf.append(getExpressionSetSummary(this.hours));
    buf.append("\n");
    buf.append("daysOfMonth: ");
    buf.append(getExpressionSetSummary(this.daysOfMonth));
    buf.append("\n");
    buf.append("months: ");
    buf.append(getExpressionSetSummary(this.months));
    buf.append("\n");
    buf.append("daysOfWeek: ");
    buf.append(getExpressionSetSummary(this.daysOfWeek));
    buf.append("\n");
    buf.append("lastdayOfWeek: ");
    buf.append(this.lastdayOfWeek);
    buf.append("\n");
    buf.append("nearestWeekday: ");
    buf.append(this.nearestWeekday);
    buf.append("\n");
    buf.append("NthDayOfWeek: ");
    buf.append(this.nthdayOfWeek);
    buf.append("\n");
    buf.append("lastdayOfMonth: ");
    buf.append(this.lastdayOfMonth);
    buf.append("\n");
    buf.append("years: ");
    buf.append(getExpressionSetSummary(this.years));
    buf.append("\n");
    return buf.toString();
  }

  protected String getExpressionSetSummary(Set<Integer> set) {
    if (set.contains(this.NO_SPEC)) {
      return "?";
    }
    if (set.contains(this.ALL_SPEC)) {
      return "*";
    }
    StringBuilder buf = new StringBuilder();
    Iterator itr = set.iterator();
    boolean first = true;
    while (itr.hasNext()) {
      Integer iVal = (Integer) itr.next();
      String val = iVal.toString();
      if (!first) {
        buf.append(",");
      }
      buf.append(val);
      first = false;
    }
    return buf.toString();
  }

  protected String getExpressionSetSummary(ArrayList<Integer> list) {
    if (list.contains(NO_SPEC)) {
      return "?";
    }
    if (list.contains(ALL_SPEC)) {
      return "*";
    }
    StringBuilder buf = new StringBuilder();
    Iterator itr = list.iterator();
    boolean first = true;
    while (itr.hasNext()) {
      Integer iVal = (Integer) itr.next();
      String val = iVal.toString();
      if (!first) {
        buf.append(",");
      }
      buf.append(val);
      first = false;
    }
    return buf.toString();
  }

  protected int skipWhiteSpace(int i, String s) {
    for (/**/; i < s.length() && (s.charAt(i) == ' ' || s.charAt(i) == '\t'); i++) {
      /* empty */
    }
    return i;
  }

  protected int findNextWhiteSpace(int i, String s) {
    for (/**/; i < s.length() && (s.charAt(i) != ' ' || s.charAt(i) != '\t'); i++) {
      /* empty */
    }
    return i;
  }

  protected void addToSet(int val, int end, int incr, int type)
      throws ParseException {
    TreeSet set = getSet(type);
    if ((type == 0) || (type == 1)) {
      if (((val < 0) || (val > 59) || (end > 59)) && (val != 99)) {
        throw new ParseException("Minute and Second values must be between 0 and 59", -1);
      }
    } else if (type == 2) {
      if (((val < 0) || (val > 23) || (end > 23)) && (val != 99)) {
        throw new ParseException("Hour values must be between 0 and 23", -1);
      }
    } else if (type == 3) {
      if (((val < 1) || (val > 31) || (end > 31)) && (val != 99) && (val != 98)) {
        throw new ParseException("Day of month values must be between 1 and 31", -1);
      }
    } else if (type == 4) {
      if (((val < 1) || (val > 12) || (end > 12)) && (val != 99)) {
        throw new ParseException("Month values must be between 1 and 12", -1);
      }
    } else if ((type == 5) &&
        ((val == 0) || (val > 7) || (end > 7)) && (val != 99) && (val != 98)) {
      throw new ParseException("Day-of-Week values must be between 1 and 7", -1);
    }
    if (((incr == 0) || (incr == -1)) && (val != 99)) {
      if (val != -1) {
        set.add(Integer.valueOf(val));
      } else {
        set.add(this.NO_SPEC);
      }
    } else {
      int startAt = val;
      int stopAt = end;
      if ((val == 99) && (incr <= 0)) {
        incr = 1;
        set.add(this.ALL_SPEC);
      }
      if ((type == 0) || (type == 1)) {
        if (stopAt == -1) {
          stopAt = 59;
        }
        if ((startAt == -1) || (startAt == 99)) {
          startAt = 0;
        }
      } else if (type == 2) {
        if (stopAt == -1) {
          stopAt = 23;
        }
        if ((startAt == -1) || (startAt == 99)) {
          startAt = 0;
        }
      } else if (type == 3) {
        if (stopAt == -1) {
          stopAt = 31;
        }
        if ((startAt == -1) || (startAt == 99)) {
          startAt = 1;
        }
      } else if (type == 4) {
        if (stopAt == -1) {
          stopAt = 12;
        }
        if ((startAt == -1) || (startAt == 99)) {
          startAt = 1;
        }
      } else if (type == 5) {
        if (stopAt == -1) {
          stopAt = 7;
        }
        if ((startAt == -1) || (startAt == 99)) {
          startAt = 1;
        }
      } else if (type == 6) {
        if (stopAt == -1) {
          stopAt = MAX_YEAR;
        }
        if ((startAt == -1) || (startAt == 99)) {
          startAt = 1970;
        }
      }
      for (int i = startAt; i <= stopAt; i += incr) {
        set.add(Integer.valueOf(i));
      }
    }
  }

  protected TreeSet<Integer> getSet(int type) {
    switch (type) {
      case 0:
        return this.seconds;
      case 1:
        return this.minutes;
      case 2:
        return this.hours;
      case 3:
        return this.daysOfMonth;
      case 4:
        return this.months;
      case 5:
        return this.daysOfWeek;
      case 6:
        return this.years;
      default:
        return null;
    }
  }

  protected ValueSet getValue(int v, String s, int i) {
    char c = s.charAt(i);
    StringBuilder s1 = new StringBuilder(String.valueOf(v));
    for (/**/; c >= '0' && c <= '9'; c = s.charAt(i)) {
      s1.append(c);
      if (++i >= s.length()) {
        break;
      }
    }
    ValueSet val = new ValueSet();
    val.pos = i < s.length() ? i : i + 1;
    val.value = Integer.parseInt(s1.toString());
    return val;
  }

  protected int getNumericValue(String s, int i) {
    int endOfVal = findNextWhiteSpace(i, s);
    String val = s.substring(i, endOfVal);
    return Integer.parseInt(val);
  }

  protected int getMonthNumber(String s) {
    Integer integer = (Integer) monthMap.get(s);
    if (integer == null) {
      return -1;
    }
    return integer.intValue();
  }

  protected int getDayOfWeekNumber(String s) {
    Integer integer = (Integer) dayMap.get(s);
    if (integer == null) {
      return -1;
    }
    return integer.intValue();
  }

  protected Date getTimeAfter(Date afterTime) {
    Calendar cl = new GregorianCalendar(getTimeZone());
    afterTime = new Date(afterTime.getTime() + 1000L);
    cl.setTime(afterTime);
    cl.set(14, 0);
    boolean gotOne = false;
    while (!gotOne) {
      if (cl.get(1) > 2999) {
        return null;
      }
      SortedSet st = null;
      int t = 0;
      int sec = cl.get(13);
      int min = cl.get(12);
      st = this.seconds.tailSet(Integer.valueOf(sec));
      if ((st != null) && (st.size() != 0)) {
        sec = ((Integer) st.first()).intValue();
      } else {
        sec = ((Integer) this.seconds.first()).intValue();
        min++;
        cl.set(12, min);
      }
      cl.set(13, sec);
      min = cl.get(12);
      int hr = cl.get(11);
      t = -1;
      st = this.minutes.tailSet(Integer.valueOf(min));
      if (st != null && st.size() != 0) {
        t = min;
        min = ((Integer) st.first()).intValue();
      } else {
        min = ((Integer) this.minutes.first()).intValue();
        hr++;
      }
      if (min != t) {
        cl.set(13, 0);
        cl.set(12, min);
        setCalendarHour(cl, hr);
      } else {
        cl.set(12, min);
        hr = cl.get(11);
        int day = cl.get(5);
        t = -1;
        st = this.hours.tailSet(Integer.valueOf(hr));
        if (st != null && st.size() != 0) {
          t = hr;
          hr = ((Integer) st.first()).intValue();
        } else {
          hr = ((Integer) this.hours.first()).intValue();
          day++;
        }
        if (hr != t) {
          cl.set(13, 0);
          cl.set(12, 0);
          cl.set(5, day);
          setCalendarHour(cl, hr);
        } else {
          cl.set(11, hr);
          day = cl.get(5);
          int mon = cl.get(2) + 1;
          t = -1;
          int tmon = mon;
          boolean dayOfMSpec = !this.daysOfMonth.contains(this.NO_SPEC);
          boolean dayOfWSpec = !this.daysOfWeek.contains(this.NO_SPEC);
          if ((dayOfMSpec) && (!dayOfWSpec)) {
            st = this.daysOfMonth.tailSet(Integer.valueOf(day));
            if (this.lastdayOfMonth) {
              if (!this.nearestWeekday) {
                t = day;
                day = getLastDayOfMonth(mon, cl.get(1));
              } else {
                t = day;
                day = getLastDayOfMonth(mon, cl.get(1));
                Calendar tcal = Calendar.getInstance(getTimeZone());
                tcal.set(13, 0);
                tcal.set(12, 0);
                tcal.set(11, 0);
                tcal.set(5, day);
                tcal.set(2, mon - 1);
                tcal.set(1, cl.get(1));
                int ldom = getLastDayOfMonth(mon, cl.get(1));
                int dow = tcal.get(7);
                if ((dow == 7) && (day == 1)) {
                  day += 2;
                } else if (dow == 7) {
                  day--;
                } else if ((dow == 1) && (day == ldom)) {
                  day -= 2;
                } else if (dow == 1) {
                  day++;
                }
                tcal.set(13, sec);
                tcal.set(12, min);
                tcal.set(11, hr);
                tcal.set(5, day);
                tcal.set(2, mon - 1);
                Date nTime = tcal.getTime();
                if (nTime.before(afterTime)) {
                  day = 1;
                  mon++;
                }
              }
            } else if (this.nearestWeekday) {
              t = day;
              day = ((Integer) this.daysOfMonth.first()).intValue();
              Calendar tcal = Calendar.getInstance(getTimeZone());
              tcal.set(13, 0);
              tcal.set(12, 0);
              tcal.set(11, 0);
              tcal.set(5, day);
              tcal.set(2, mon - 1);
              tcal.set(1, cl.get(1));
              int ldom = getLastDayOfMonth(mon, cl.get(1));
              int dow = tcal.get(7);
              if ((dow == 7) && (day == 1)) {
                day += 2;
              } else if (dow == 7) {
                day--;
              } else if ((dow == 1) && (day == ldom)) {
                day -= 2;
              } else if (dow == 1) {
                day++;
              }
              tcal.set(13, sec);
              tcal.set(12, min);
              tcal.set(11, hr);
              tcal.set(5, day);
              tcal.set(2, mon - 1);
              Date nTime = tcal.getTime();
              if (nTime.before(afterTime)) {
                day = ((Integer) this.daysOfMonth.first()).intValue();
                mon++;
              }
            } else if (st != null && st.size() != 0) {
              t = day;
              day = ((Integer) st.first()).intValue();
            } else {
              day = ((Integer) this.daysOfMonth.first()).intValue();
              mon++;
            }
            if ((day != t) || (mon != tmon)) {
              cl.set(13, 0);
              cl.set(12, 0);
              cl.set(11, 0);
              cl.set(5, day);
              cl.set(2, mon - 1);
              continue;
            }
          } else if ((dayOfWSpec) && (!dayOfMSpec)) {
            if (this.lastdayOfWeek) {
              int dow = ((Integer) this.daysOfWeek.first()).intValue();
              int cDow = cl.get(7);
              int daysToAdd = 0;
              if (cDow < dow) {
                daysToAdd = dow - cDow;
              }
              if (cDow > dow) {
                daysToAdd = dow + (7 - cDow);
              }
              int lDay = getLastDayOfMonth(mon, cl.get(1));
              if (day + daysToAdd > lDay) {
                cl.set(13, 0);
                cl.set(12, 0);
                cl.set(11, 0);
                cl.set(5, 1);
                cl.set(2, mon);
                continue;
              }
              for (/**/; day + daysToAdd + 7 <= lDay; daysToAdd += 7) {
                /* empty */
              }
              day += daysToAdd;
              if (daysToAdd > 0) {
                cl.set(13, 0);
                cl.set(12, 0);
                cl.set(11, 0);
                cl.set(5, day);
                cl.set(2, mon - 1);
                continue;
              }
            } else if (this.nthdayOfWeek != 0) {
              int dow = ((Integer) this.daysOfWeek.first()).intValue();
              int cDow = cl.get(7);
              int daysToAdd = 0;
              if (cDow < dow) {
                daysToAdd = dow - cDow;
              } else if (cDow > dow) {
                daysToAdd = dow + (7 - cDow);
              }
              boolean dayShifted = false;
              if (daysToAdd > 0) {
                dayShifted = true;
              }
              day += daysToAdd;
              int weekOfMonth = day / 7;
              if (day % 7 > 0) {
                weekOfMonth++;
              }
              daysToAdd = (this.nthdayOfWeek - weekOfMonth) * 7;
              day += daysToAdd;
              if ((daysToAdd < 0) || (day > getLastDayOfMonth(mon, cl.get(1)))) {
                cl.set(13, 0);
                cl.set(12, 0);
                cl.set(11, 0);
                cl.set(5, 1);
                cl.set(2, mon);
                continue;
              }
              if ((daysToAdd > 0) || (dayShifted)) {
                cl.set(13, 0);
                cl.set(12, 0);
                cl.set(11, 0);
                cl.set(5, day);
                cl.set(2, mon - 1);
                continue;
              }
            } else {
              int cDow = cl.get(7);
              int dow = ((Integer) this.daysOfWeek.first()).intValue();
              st = this.daysOfWeek.tailSet(Integer.valueOf(cDow));
              if (st != null && st.size() > 0) {
                dow = ((Integer) st.first()).intValue();
              }
              int daysToAdd = 0;
              if (cDow < dow) {
                daysToAdd = dow - cDow;
              }
              if (cDow > dow) {
                daysToAdd = dow + (7 - cDow);
              }
              int lDay = getLastDayOfMonth(mon, cl.get(1));
              if (day + daysToAdd > lDay) {
                cl.set(13, 0);
                cl.set(12, 0);
                cl.set(11, 0);
                cl.set(5, 1);
                cl.set(2, mon);
                continue;
              }
              if (daysToAdd > 0) {
                cl.set(13, 0);
                cl.set(12, 0);
                cl.set(11, 0);
                cl.set(5, day + daysToAdd);
                cl.set(2, mon - 1);
                continue;
              }
            }
          } else {
            throw new UnsupportedOperationException(
                "Support for specifying both a day-of-week AND a day-of-month parameter is not implemented.");
          }
          cl.set(5, day);
          mon = cl.get(2) + 1;
          int year = cl.get(1);
          t = -1;
          if (year > MAX_YEAR) {
            return null;
          }
          st = this.months.tailSet(Integer.valueOf(mon));
          if (st != null && st.size() != 0) {
            t = mon;
            mon = ((Integer) st.first()).intValue();
          } else {
            mon = ((Integer) this.months.first()).intValue();
            year++;
          }
          if (mon != t) {
            cl.set(13, 0);
            cl.set(12, 0);
            cl.set(11, 0);
            cl.set(5, 1);
            cl.set(2, mon - 1);
            cl.set(1, year);
          } else {
            cl.set(2, mon - 1);
            year = cl.get(1);
            t = -1;
            st = this.years.tailSet(Integer.valueOf(year));
            if ((st != null) && (st.size() != 0)) {
              t = year;
              year = ((Integer) st.first()).intValue();
            } else {
              return null;
            }
            if (year != t) {
              cl.set(13, 0);
              cl.set(12, 0);
              cl.set(11, 0);
              cl.set(5, 1);
              cl.set(2, 0);
              cl.set(1, year);
            } else {
              cl.set(1, year);
              gotOne = true;
            }
          }
        }
      }
    }
    return cl.getTime();
  }

  protected void setCalendarHour(Calendar cal, int hour) {
    cal.set(11, hour);
    if ((cal.get(11) != hour) && (hour != 24)) {
      cal.set(11, hour + 1);
    }
  }

  protected Date getTimeBefore(Date endTime) {
    return null;
  }

  public Date getFinalFireTime() {
    return null;
  }

  protected boolean isLeapYear(int year) {
    return ((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0);
  }

  protected int getLastDayOfMonth(int monthNum, int year) {
    switch (monthNum) {
      case 1:
        return 31;
      case 2:
        return isLeapYear(year) ? 29 : 28;
      case 3:
        return 31;
      case 4:
        return 30;
      case 5:
        return 31;
      case 6:
        return 30;
      case 7:
        return 31;
      case 8:
        return 31;
      case 9:
        return 30;
      case 10:
        return 31;
      case 11:
        return 30;
      case 12:
        return 31;
      default:
        throw new IllegalArgumentException(new StringBuilder().append("Illegal month number: ").append(monthNum).toString());
    }
  }

  private void readObject(ObjectInputStream stream)
      throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    try {
      buildExpression(this.cronExpression);
    } catch (Exception exception) {
      /* empty */
    }
  }

  public Object clone() {
    CronExpression copy = null;
    try {
      copy = new CronExpression(getCronExpression());
      copy.setTimeZone(getTimeZone());
    } catch (ParseException ex) {
      throw new IncompatibleClassChangeError("Not Cloneable.");
    }
    return copy;
  }

  static {
    monthMap.put("JAN", Integer.valueOf(0));
    monthMap.put("FEB", Integer.valueOf(1));
    monthMap.put("MAR", Integer.valueOf(2));
    monthMap.put("APR", Integer.valueOf(3));
    monthMap.put("MAY", Integer.valueOf(4));
    monthMap.put("JUN", Integer.valueOf(5));
    monthMap.put("JUL", Integer.valueOf(6));
    monthMap.put("AUG", Integer.valueOf(7));
    monthMap.put("SEP", Integer.valueOf(8));
    monthMap.put("OCT", Integer.valueOf(9));
    monthMap.put("NOV", Integer.valueOf(10));
    monthMap.put("DEC", Integer.valueOf(11));
    dayMap.put("SUN", Integer.valueOf(1));
    dayMap.put("MON", Integer.valueOf(2));
    dayMap.put("TUE", Integer.valueOf(3));
    dayMap.put("WED", Integer.valueOf(4));
    dayMap.put("THU", Integer.valueOf(5));
    dayMap.put("FRI", Integer.valueOf(6));
    dayMap.put("SAT", Integer.valueOf(7));
  }
}
