package com.brilliance.eibs.util;

import com.brilliance.eibs.core.exception.InterfaceException;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.filefilter.*;

import java.io.*;
import java.util.*;

/**
 * 文件操作工具类
 * @author xiaoyuanzhen
 *
 */
public class FileUtil
{

  public static List<String> readAsList(String filename, int begLine, int lineCount)
      throws IOException
  {
    File file = new File(filename);
    if (!file.exists() || !file.isFile())
      return null;
    if (begLine < 1 || lineCount < 0)
      return null;
    BufferedReader br = null;
    br = new BufferedReader(new FileReader(file));
    int count = 1;
    while (count++ < begLine)
      br.readLine();
    count = 1;
    String line = null;
    List<String> lines = new ArrayList<String>();
    while (count++ <= lineCount && (line = br.readLine()) != null)
    {

      lines.add(line);
    }
    br.close();
    return lines;
  }

  /**
   * 按照文件最后的修改时间排序
   * 
   * @param directory 目录
   * @param regex 过滤条件
   * @return
   */
  @SuppressWarnings("unchecked")
  public static File[] fileLastModifiedSort(String directory, final String regex)
  {
    File dir = new File(directory);
    if (dir.isDirectory())
    {
      File[] files = dir.listFiles(new FilenameFilter()
      {

        @Override
        public boolean accept(File file, String name)
        {
          if (null == regex)
            return true;
          return name.matches(regex);
        }
      });
      Arrays.sort(files, new LastModifiedFileComparator());
      return files;
    }
    return null;
  }

  /**
   * 按照文件最后的修改时间排序（整个目录）
   * 
   * @param directory
   * @return
   */
  public static File[] fileLastModifiedSort(String directory)
  {
    return fileLastModifiedSort(directory, null);
  }

  /**
   * 用来比较两个文件的最后修改时间
   * 
   * @author hujun
   *
   */
  @SuppressWarnings("rawtypes")
  static class LastModifiedFileComparator
      implements Comparator
  {
    @Override
    public int compare(Object object1, Object object2)
    {
      File file1 = (File) object1;
      File file2 = (File) object2;
      long result = file1.lastModified() - file2.lastModified();
      if (result < 0)
      {
        return -1;
      }
      else if (result > 0)
      {
        return 1;
      }
      else
      {
        return 0;
      }
    }
  }

  /** 
   * 复制文件或者目录,复制前后文件完全一样。 
   * @param resFilePath 源文件路径 
   * @param distFolder    目标文件夹
   * @return boolean  
   */
  public static boolean copyFile(String resFilePath, String distFolder)
  {
    boolean copy = false;
    try
    {
      File resFile = new File(resFilePath);
      File distFile = new File(distFolder);
      if (resFile.isDirectory())
      {
        FileUtils.copyDirectoryToDirectory(resFile, distFile);
      }
      else if (resFile.isFile())
      {
        FileUtils.copyFileToDirectory(resFile, distFile, true);
      }
      copy = true;
    }
    catch (Exception e)
    {
      throw new InterfaceException("00411", "file copy error:", e);
    }
    return copy;
  }

  /** 
    * 删除一个文件或者目录 
    * @param targetPath 文件或者目录路径 
   */
  public static boolean deleteFile(String targetPath)
  {
    boolean delete = false;
    try
    {
      File targetFile = new File(targetPath);
      if (targetFile.isDirectory())
      {
        FileUtils.deleteDirectory(targetFile);
      }
      else if (targetFile.isFile())
      {
        targetFile.delete();
      }
      delete = true;
    }
    catch (Exception e)
    {
      delete = false;
    }
    return delete;
  }

  /** 
   *过滤 删除包含某个字符串的文件 
   * @param targetPath 文件或者目录路径
   * @param containstr 文件名包含的字符串 
  */
  public static boolean deleteFileFilter(String targetPath, String containstr)
  {
    boolean delete = true;
    try
    {
      File targetFile = new File(targetPath);
      if (targetFile.isDirectory())
      {
        File file[] = targetFile.listFiles();
        for (int i = 0; i < file.length; i++)
        {
          String filename = file[i].getName();
          if (filename.contains(containstr))
          {
            file[i].delete();
          }
        }

      }
    }
    catch (Exception e)
    {
      delete = false;
    }
    return delete;
  }

  /** 
    * 移动文件或者目录,移动前后文件完全一样,如果目标文件夹不存在则创建。 
    * @param resFilePath 源文件路径 
   * @param distFolder    目标文件夹 
   * @return boolean
   */
  public static boolean moveFile(String from, String to)
  {
    boolean move = false;
    File dir = new File(from);
    if (!dir.exists())
    {
      return false;
    }
    // 文件一览 
    File[] files = dir.listFiles();
    if (dir.isDirectory())
    {
      try
      {
        if (files.length == 0 || files == null)
          return move;
        // 目标 
        File moveDir = new File(to);
        if (!moveDir.exists())
        {
          moveDir.mkdirs();
        }
        // 文件移动 
        for (int i = 0; i < files.length; i++)
        {
          if (files[i].isDirectory())
          {
            moveFile(files[i].getPath(), to + "/" + files[i].getName());
            // 成功，删除原文件 
            files[i].delete();
          }
          File moveFile = new File(moveDir.getPath() + "/" + files[i].getName());
          // 目标文件夹下存在的话，删除 
          if (moveFile.exists())
          {
            moveFile.delete();
          }
          files[i].renameTo(moveFile);
        }
        move = true;
      }
      catch (Exception e)
      {
        return move;
      }
    }
    else
    {
      // 目标 
      File moveDir = new File(to);
      if (!moveDir.exists())
      {
        moveDir.mkdirs();
      }
      String fileName = dir.getName();
      File moveFile = new File(moveDir.getPath() + "/" + fileName);
      // 目标文件夹下存在的话，删除 
      if (moveFile.exists())
      {
        moveFile.delete();
      }
      dir.renameTo(moveFile);
      move = true;
    }
    return move;
  }

  /** 
   * 移动文件或者目录,移动前后文件完全一样,如果目标文件夹不存在则创建。 
   * @param resFilePath 源文件路径 
  * @param distFolder    目标文件夹 
  * @param continstr      文件名包含的字符串
  * @return boolean
  */
  public static boolean moveFileFilter(String from, String to, String containstr)
  {
    boolean move = false;
    File dir = new File(from);
    if (!dir.exists())
    {
      return false;
    }
    // 文件一览 
    File[] files = dir.listFiles();
    if (dir.isDirectory())
    {
      try
      {
        if (files.length == 0 || files == null)
          return move;
        // 目标 
        File moveDir = new File(to);
        if (!moveDir.exists())
        {
          moveDir.mkdirs();
        }
        // 文件移动 
        for (int i = 0; i < files.length; i++)
        {
          String filename = files[i].getName();
          boolean iscontain = filename.contains(containstr);
          if (files[i].isDirectory() && iscontain)
          {
            moveFileFilter(files[i].getPath(), to + "/" + files[i].getName(), containstr);
            // 成功，删除原文件 
            files[i].delete();
          }
          File moveFile = new File(moveDir.getPath() + "/" + files[i].getName());
          // 目标文件夹下存在的话，删除 
          if (moveFile.exists())
          {
            moveFile.delete();
          }
          if (iscontain)
          {
            files[i].renameTo(moveFile);
            iscontain = false;
          }
        }
        move = true;
      }
      catch (Exception e)
      {
        return move;
      }
    }
    else
    {
      // 目标 
      File moveDir = new File(to);
      if (!moveDir.exists())
      {
        moveDir.mkdirs();
      }
      String fileName = dir.getName();
      File moveFile = new File(moveDir.getPath() + "/" + fileName);
      // 目标文件夹下存在的话，删除 
      if (moveFile.exists())
      {
        moveFile.delete();
      }
      if (fileName.contains(containstr))
      {
        dir.renameTo(moveFile);
      }
      move = true;
    }
    return move;
  }

  /** 
    * 读取文件或者目录的大小 
   * 
    * @param distFilePath 目标文件或者文件夹 
   * @return 文件或者目录的大小，如果获取失败，则返回-1 
    */
  public long genFileSize(String distFilePath)
  {
    File distFile = new File(distFilePath);
    if (distFile.isFile())
    {
      return distFile.length();
    }
    else if (distFile.isDirectory())
    {
      return FileUtils.sizeOfDirectory(distFile);
    }
    return -1L;
  }

  /** 
    * 判断一个文件是否存在 
   * 
    * @param filePath 文件路径 
   * @return 存在返回true，否则返回false 
    */
  public static boolean isExist(String filePath)
  {
    return new File(filePath).exists();
  }

  /** 
    * 本地某个目录下的文件列表（不递归） 
   * 
    * @param folder  某个目录 
   * @param suffix 某个目录下文件的后缀名（比如.mov.xml) 
    * @return 文件名称列表 String[]
   */
  public String[] listFilebySuffix(String folder, String suffix)
  {
    IOFileFilter fileFilter1 = new SuffixFileFilter(suffix);
    IOFileFilter fileFilter2 = new NotFileFilter(DirectoryFileFilter.INSTANCE);
    FilenameFilter filenameFilter = new AndFileFilter(fileFilter1, fileFilter2);
    return new File(folder).list(filenameFilter);
  }

  /** 
    * 将字符串写入指定文件(当指定的父路径中文件夹不存在时，会最大限度去创建，以保证保存成功！) 
    * 
    * @param res       原字符串 
   * @param filePath 文件路径 
   * @return 成功标记 
   */
  public boolean string2File(String res, String filePath)
  {
    boolean flag = true;
    BufferedReader bufferedReader = null;
    BufferedWriter bufferedWriter = null;
    try
    {
      File distFile = new File(filePath);
      if (!distFile.getParentFile().exists())
        distFile.getParentFile().mkdirs();
      bufferedReader = new BufferedReader(new StringReader(res));
      bufferedWriter = new BufferedWriter(new FileWriter(distFile));
      char buf[] = new char[1024]; //字符缓冲区 
      int len;
      while ((len = bufferedReader.read(buf)) != -1)
      {
        bufferedWriter.write(buf, 0, len);
      }
      bufferedWriter.flush();
      bufferedReader.close();
      bufferedWriter.close();
    }
    catch (IOException e)
    {
      flag = false;
      throw new InterfaceException("00414", "string write to file error.", e);
    }
    return flag;
  }

  /**
   * @param 报文内容
   * @param record 换行符
   * @return 报文行数
   * **/
  public int getFileLines(String str, String record)
  {
    int lines = 0;
    String strs[] = str.split(record);
    for (int i = 0; i < strs.length; i++)
    {
      if (!StringUtil.isEmpty(strs[i]))
      {
        lines++;
      }
    }
    return lines;
  }

  /**
   * 获取文件名,不包含路径,不包含后缀
   * 
   * @param filename
   */
  public static String getFileName(String filename)
  {
    String name = filename;
    int index1 = filename.lastIndexOf('/');
    int index2 = filename.lastIndexOf('\\');
    int index = index1 > index2 ? index1 : index2;
    if (index != -1)
    {
      name = filename.substring(index + 1);
    }
    index = name.lastIndexOf(".");
    if (index != -1)
      return name.substring(0, index);
    return name;
  }

  //public static List<String> list = new ArrayList<String>();

  public static void getFileNamesByRegexList(String path, String regex, List<String> list)
  {
    File filedir = new File(path);
    for (File item : filedir.listFiles())
    {
      if (item.isDirectory())
      {
        getFileNamesByRegexList(item.getPath(), regex, list);
      }
      else if (item.getName().matches(regex))
      {
        list.add(item.getName());
      }
    }
  }

  public static List<String> getFileNamesByRegex(String path, String regex)
  {
    List<String> list = new ArrayList<String>();
    getFileNamesByRegexList(path, regex, list);
    return list;

  }

  
  public static Collection<File>  getFiles(File dir)
  {
    Collection<File> files = FileUtils.listFiles(dir, FileFilterUtils.suffixFileFilter(".xml"),
        DirectoryFileFilter.INSTANCE);
    return files;
  }
  
 
  
  public static void main(String[] args) throws FileNotFoundException, IOException
  {

    Collection<File> files = FileUtils.listFiles(new File("C:/Users/ge/Desktop/src/"), FileFilterUtils.suffixFileFilter("txt"),
        DirectoryFileFilter.INSTANCE);
    Iterator<File> it = files.iterator();
    int a=0;
    while(it.hasNext())
    {
      
      File f=it.next();
      System.out.println(f.getPath());
     List<String> list=IOUtils.readLines(new FileInputStream(f));
     a=a+list.size();
    }
    System.out.println(a);
    
   // String a="d:/a/cfg/main/a/1.xml";
 //   System.out.println(a.substring(a.indexOf("cfg/main")));
 
  
  }
  

 

}
