package com.brilliance.eibs.util.logging;


import org.slf4j.bridge.SLF4JBridgeHandler;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.logging.ConsoleHandler;
import java.util.logging.Handler;
import java.util.logging.LogManager;
import java.util.logging.Logger;

public abstract class AbstractLoggingFrame extends LoggingFrame {

    private static final String BRIDGE_HANDLER = "org.slf4j.bridge.SLF4JBridgeHandler";
    private final ClassLoader classLoader;
    protected String configLocation;

    public AbstractLoggingFrame(ClassLoader classLoader) {
        this.classLoader = classLoader;
        this.configLocation = getConfigPath();
    }

    protected static boolean isPresent(String className, ClassLoader classLoader) {
        try {
            Class.forName(className, false, classLoader);
            return true;
        } catch (IllegalAccessError err) {
            throw new IllegalStateException("Readability mismatch in inheritance hierarchy of class [" +
                    className + "]: " + err.getMessage(), err);
        } catch (Throwable ex) {
            // Typically ClassNotFoundException or NoClassDefFoundError...
            return false;
        }
    }

    public static URL getURL(String resourceLocation) {
        try {
            //重复:考虑有的没有加classpath:前缀
            ClassLoader cl = LoggingFrame.class.getClassLoader();
            URL url = (cl != null ? cl.getResource(resourceLocation) : ClassLoader.getSystemResource(resourceLocation));
            if (url == null) {
                //视为url协议
                return new URL(resourceLocation);
            }
            return url;
        } catch (MalformedURLException ex) {
            // 视为文件路径
            try {
                return new File(resourceLocation).toURI().toURL();
            } catch (MalformedURLException ex2) {
                throw new IllegalArgumentException("Resource location [" + resourceLocation +
                        "] is neither a URL not a well-formed file path");
            }
        }
    }

    public static File getFile(URL resourceUrl) {
        if (!"file".equals(resourceUrl.getProtocol())) {
            throw new IllegalArgumentException("url cannot be resolved to absolute file path " +
                    "because it does not reside in the file system: " + resourceUrl);
        }
        return new File(toURI(resourceUrl).getSchemeSpecificPart());
    }

    public static URI toURI(URL url) {
        return toURI(url.toString());
    }

    public static URI toURI(String location) {
        try {
            return new URI(location.replaceAll(" ", "%20"));
        } catch (Exception e) {
            throw new IllegalArgumentException("URI转换异常:" + e.getMessage(), e);
        }

    }

    protected abstract void loadConfiguration(String location);

    protected abstract String[] getStandardConfigLocations();

    @Override
    public void beforeInitialize() {
        configureJdkLoggingBridgeHandler();
    }

    protected void reinitialize(String configLocation) {
    }

    protected abstract String getConfigPath();

    @Override
    public void initialize() {
        loadConfiguration(configLocation);
    }

    protected final ClassLoader getClassLoader() {
        return this.classLoader;
    }

    private void configureJdkLoggingBridgeHandler() {
        try {
            if (isBridgeJulIntoSlf4j()) {
                removeJdkLoggingBridgeHandler();
                SLF4JBridgeHandler.install();
            }
        } catch (Throwable ex) {
            // 静默处理. java.util.logging bridge 没有被加载.
        }
    }

    protected final boolean isBridgeJulIntoSlf4j() {
        return isBridgeHandlerAvailable() && isJulUsingASingleConsoleHandlerAtMost();
    }

    protected final boolean isBridgeHandlerAvailable() {
        return isPresent(BRIDGE_HANDLER, getClassLoader());
    }

    private boolean isJulUsingASingleConsoleHandlerAtMost() {
        Logger rootLogger = LogManager.getLogManager().getLogger("");
        Handler[] handlers = rootLogger.getHandlers();
        return handlers.length == 0 || (handlers.length == 1 && handlers[0] instanceof ConsoleHandler);
    }

    private void removeJdkLoggingBridgeHandler() {
        try {
            removeDefaultRootHandler();
            SLF4JBridgeHandler.uninstall();
        } catch (Throwable ex) {
        }
    }

    private void removeDefaultRootHandler() {
        try {
            Logger rootLogger = LogManager.getLogManager().getLogger("");
            Handler[] handlers = rootLogger.getHandlers();
            if (handlers.length == 1 && handlers[0] instanceof ConsoleHandler) {
                rootLogger.removeHandler(handlers[0]);
            }
        } catch (Throwable ex) {
        }
    }
}
